/**
  * Autor: Victor Carbune
  * Echipa SD, 2012
  *
  * Modificari: Adrian Bogatu
  * Echipa SD, 2013
  *
  * Modificari: Cosmin Petrisor
  * Echipa SD, 2015
  *
  * Modificari: Cristian Creteanu
  * Echipa SD, 2017
  * 
  * Modificari: Ioana Stefan
  * Echipa SD, 2018
  * License: LGPL
  */

#include <iostream>
#include <fstream>
#include <vector>
#include <stdlib.h>
#include <time.h>
#include <cassert>

#include "BinarySearchTree.h"
#include "BinaryTree.h"
#include "Heap.h"

void baseFuntions(int nNumbers, std::vector<int> numbers, BinarySearchTree<int> *root) {
    std::ofstream task1("../Out/Lab9/task1.out");

    int key;
    for (int i=0;i<2;i++)
    {
        int pos=nNumbers - i*2 - 1;
        key=numbers[pos];
        if (root->searchKey(key)) {
            std::cout<<"Element "<<key<<" found"<<std::endl;
            task1<<"Element "<<key<<" found"<<std::endl;
        } else {
            std::cout<<"[ERROR]Element "<<key<<" not found"<<std::endl;
            task1<<"[ERROR]Element "<<key<<" not found"<<std::endl;
        }
    }
    key=100;
    if (root->searchKey(key)) {
        std::cout<<"Element "<<key<<" found"<<std::endl;
        task1<<"Element "<<key<<" found"<<std::endl;
    } else {
        std::cout<<"Element "<<key<<" not found"<<std::endl;
        task1<<"Element "<<key<<" not found"<<std::endl;
    }
    
    std::vector<int> display = root->inOrderDisplay();
    for (int elem : display) {
        std::cout<<elem<<" ";
        task1<<elem<<" ";
    }
    std::cout<<std::endl;
}

void findMinMax(BinarySearchTree<int> *root) {
    std::ofstream task2("../Out/Lab9/task2.out");

    std::cout<<"Min element: "<<root->findMin()<<std::endl;
    std::cout<<"Max element: "<<root->findMax()<<std::endl;
    task2<<"Min element: "<<root->findMin()<<std::endl;
    task2<<"Max element: "<<root->findMax()<<std::endl;
    std::cout << std::endl;
}

void findDisplayLevels(BinarySearchTree<int> *root) {
    std::ofstream task3("../Out/Lab9/task3.out");

    int nLevels=root->findLevels();
    std::cout<<"Tree has "<<nLevels<<" levels"<<std::endl;
    task3<<"Tree has "<<nLevels<<" levels"<<std::endl;
    for (int i=0;i<nLevels;i++) {
        std::vector<int> dl = root->displayLevel(i);
        for (int elem : dl) {
            std::cout<<elem<<" ";
            task3<<elem<<" ";
        }
        std::cout<<std::endl;
        task3<<std::endl;
    }
    std::cout<<std::endl;
}

void removeGivenKeys(std::vector<int> numbers, BinarySearchTree<int> *root) {
    std::ofstream task4("../Out/Lab9/task4.out");

    int key;
    for (int i=0;i<2;i++)
    {
        int pos=i+2;
        key=numbers[pos];
        root=root->removeKey(key);
        if (root->searchKey(key)) {
            std::cout<<"[ERROR]Element "<<key<<" found. "<<std::endl;
            task4<<"[ERROR]Element "<<key<<" found. "<<std::endl;
        } else {
            std::cout<<"Element "<<key<<" not found"<<std::endl;
            task4<<"Element "<<key<<" not found"<<std::endl;
        }
        std::cout<<"InOrder Display: ";
        std::vector<int> display = root->inOrderDisplay();
        for (int elem : display) {
            std::cout<<elem<<" ";
            task4<<elem<<" ";
        }
        std::cout<<std::endl;
        task4<<std::endl;
    }
    std::cout << std::endl;
}

void checkIsBST() {
    std::ofstream task5("../Out/Lab9/task5.out");

    auto *root2 = new BinaryTree<int>();
    root2->setData(1);
    root2->insertLeft(2);
    root2->insertRight(3);
    auto rootLeftSubtree = root2->getLeftSubtree();
    rootLeftSubtree->insertLeft(4);
    rootLeftSubtree->insertRight(5);
    auto rootRightSubtree = root2->getRightSubtree();
    rootRightSubtree->insertRight(7);
    auto leftLeftSubtree = rootLeftSubtree->getLeftSubtree();
    leftLeftSubtree->insertLeft(10);
    leftLeftSubtree->insertRight(8);

    if (!(BinaryTree<int>::isBST(root2))) {
        std::cout<<"Passed"<<std::endl;
        task5<<"Passed"<<std::endl;
    } else {
        std::cout<<"Failed"<<std::endl;
        task5<<"Failed"<<std::endl;
    }
    delete root2;

    auto *root3 = new BinaryTree<int>();
    root3->setData(7);
    root3->insertLeft(3);
    root3->insertRight(15);

    root3->getLeftSubtree()->insertLeft(1);
    root3->getLeftSubtree()->insertRight(4);

    root3->getRightSubtree()->insertLeft(9);
    root3->getRightSubtree()->insertRight(18);

    assert(BinaryTree<int>::isBST(root3));
    if (BinaryTree<int>::isBST(root3)) {
        std::cout<<"Passed"<<std::endl;
        task5<<"Passed"<<std::endl;
    } else {
        std::cout<<"Failed"<<std::endl;
        task5<<"Failed"<<std::endl;
    }
    delete root3;
    std::cout << std::endl;
}

void checkHeap() {
    std::ofstream bonus1("../Out/Lab9/bonus1.out");

    Heap<int> heap(10);

    heap.insert(23);
    heap.insert(12);
    heap.insert(76);

    std::cout << heap.peek() << std::endl;
    bonus1 << heap.peek() << std::endl;
    heap.extractMin();
    std::cout << heap.peek() << std::endl;
    bonus1 << heap.peek() << std::endl;
}

int main(int argc,char** argv)
{
    if (argc!=2)
    {
        std::cerr<<"Usage: "<<argv[0]<<" inputFile"<<std::endl;
        exit(0);
    }
    // Open file
    std::ifstream in(argv[1],std::ios::in);
    if (!in.is_open())
    {
        std::cerr<<"Invalid input file:"<<argv[0]<<std::endl;
        exit(0);
    }
    
    // Read data from file
    int nNumbers;
    in>>nNumbers;
    std::vector<int> numbers;
    for (int i=0;i<nNumbers;i++)
    {
        int nReadNumber;
        in>>nReadNumber;
        numbers.push_back(nReadNumber);
    }
    in.close();
    
    // Insert numbers in BST
    BinarySearchTree<int> *root=new BinarySearchTree<int>();
    for (std::vector<int>::iterator it=numbers.begin();it!=numbers.end();it++)
    {
        root->insertKey(*it);
    }

    baseFuntions(nNumbers, numbers, root);
    findMinMax(root);
    findDisplayLevels(root);
    removeGivenKeys(numbers, root);
    delete root;

    checkIsBST();
    checkHeap();

    return 0;
}
